import simpy
import pandas as pd
import numpy as np
import datetime
import time
from scipy import stats
from collections import namedtuple
import seaborn as sns
import matplotlib.pyplot as plt


class Simulation:
    """
    A discrete event simulation that simulates the queue.
    - queue is a queue based on the queue class
    - seed is a random seed to have retraceable simulations
    """

    def __init__(self, queue, max_arr=100, priority=False, seed=None):
        """
        Initialization (the basic time unit is hours)
        """

        self.queue = queue
        self.max_arr = max_arr

        # set simulation time and epoch
        self.sim_start = datetime.datetime.now()
        self.env = simpy.Environment(initial_time=time.mktime(self.sim_start.timetuple()))
        self.env.epoch = time.mktime(self.sim_start.timetuple())

        # initialise counters and logs
        self.c_s = 0  # people in the system
        self.c_q = 0  # people in the queue
        self.system_state = {
            "t": [0],
            "c_s": [0],
            "c_q": [0]}

        self.customer_nr = 0
        self.log = {
            "c_id": [],  # c_id = customer id
            "IAT": [],  # IAT = inter arrival time
            "ST": [],  # ST = service time
            "AT": [],  # AT = now + IAT
            "TSB": [],  # TSB = time service begins
            "TSE": [],  # TSE = time service ends
            "TCSS": [],  # TCSS = time customer spends in the system
            "TCWQ": [],  # TCWQ = time customer waits in the queue
            "ITS": [],  # ITS = idle time of the server
            "s_id": []}  # s_id = server id

        # activate random seed
        np.random.seed(seed)

        # define arrival and service processes
        if not priority:

            # --- arrival distribution ---
            if self.queue.A.symbol == "M":
                # define the average inter arrival time and add distribution with appropriate scaling
                aver_IAT = 1 / self.queue.A.arr_rate
                self.queue.A.arrival_distribution = stats.expon(scale=aver_IAT)

            elif self.queue.A.symbol[0] == "E":
                # define the average inter arrival time and add distribution with appropriate scaling
                aver_IAT = 1 / self.queue.A.arr_rate
                k = int(self.queue.A.symbol[1:])
                loc = 0
                self.queue.A.arrival_distribution = stats.erlang(k, loc=loc, scale=aver_IAT / k)

            elif self.queue.A.symbol == "D":
                # the deterministic type expects arr_rate to contain a dataframe with columns ["name","IAT","AT"]
                self.queue.A.arrival_distribution = self.queue.A.arr_rate

            # --- service distribution ---
            self.env.servers = simpy.FilterStore(self.env, capacity=self.queue.c)
            self.env.servers.items = []  # to be filled in the next steps depending on S.symbol
            self.env.server_info = {}  # to be filled in the next steps depending on S.symbol
            Server = namedtuple('Server', 'service_distribution, last_active, id')

            if self.queue.S.symbol == "M":
                # define the average service time and add distribution with appropriate scaling for each server
                aver_ST = 1 / self.queue.S.srv_rate
                for i in range(1, self.queue.c + 1):
                    self.env.servers.items.append(Server(stats.expon(scale=aver_ST), self.env.now, i))
                    self.env.server_info.update({i: {'last_active': self.env.now}})

            elif self.queue.S.symbol[0] == "E":
                # define the average service time and add distribution with appropriate scaling for each server
                aver_ST = 1 / self.queue.S.srv_rate
                for i in range(1, self.queue.c + 1):
                    k = int(self.queue.S.symbol[1:])
                    loc = 0
                    self.env.servers.items.append(Server(stats.erlang(k, loc=loc,  scale=aver_ST/k), self.env.now, i))
                    self.env.server_info.update({i: {'last_active': self.env.now}})

            elif self.queue.S.symbol == "D":
                if self.queue.c == 1:
                    # for 1 server the deterministic type expects srv_rate to contain a dataframe
                    # with columns: ["name","ST"]
                    self.env.servers.items.append(Server(self.queue.S.srv_rate, self.env.now, 1))
                    self.env.server_info.update({1: {'last_active': self.env.now}})
                else:
                    # for n servers the deterministic type expects srv_rate to contain a dataframe
                    # with columns: ["name","ST","s_id"]
                    for i in range(1, self.queue.c + 1):
                        self.env.servers.items.append(Server(self.queue.S.srv_rate[self.queue.S.srv_rate["s_id"] == i],
                                                             self.env.now, i))
                        self.env.server_info.update({i: {'last_active': self.env.now}})

        else:
            pass
            # Todo: add the option of having priority arrivals?

        # initiate queue populating process
        self.env.process(self.queue.populate(self.env, self))

    def run(self, max_arr=1000):
        """
        Run simulation
        """

        self.max_arr = max_arr

        self.env.run()

    def log_customer_state(self, customer_id, IAT, AT, ST, TSB, TSE, ITS, s_id):
        """
        # the following items are logged per customer that enters the system:
        # c = customer id
        # IAT = inter arrival time
        # ST = service time
        # AT = arrival time
        # TSB = time service begins
        # TSE = time service ends
        # TCSS = time customer spends in the system
        # TCWQ = time customer waits in the queue
        # ITS = idle time of the server
        # s_id = id of server assigned to customer
        """

        self.log["c_id"].append(customer_id)
        self.log["IAT"].append(IAT)
        self.log["ST"].append(ST)
        self.log["AT"].append(AT)
        self.log["TSB"].append(TSB)
        self.log["TSE"].append(TSE)
        self.log["TCWQ"].append(TSB - AT)
        self.log["TCSS"].append(TSE - AT)
        self.log["ITS"].append(ITS)
        self.log["s_id"].append(s_id)

    def log_system_state(self, t, c_s, c_q):
        """
        # the following items are logged for the state of the system:
        # t = time (from start of simulation)
        # c_s = number of customers in the system
        # c_q = number of customers in the queue
        """

        self.system_state["t"].append(t)
        self.system_state["c_s"].append(c_s)
        self.system_state["c_q"].append(c_q)

    def return_log(self):
        """
        Return the log in the form of a pandas data frame.
        """

        # convert self.log to dataframe
        df_cust = pd.DataFrame.from_dict(self.log)
        df_cust = df_cust.sort_values(by=['AT'], ascending=[True])

        # convert self.system_state to dataframe
        df_sys = pd.DataFrame.from_dict(self.system_state)
        df_sys = df_sys.sort_values(by=['t'], ascending=[True])

        return df_cust, df_sys

    def get_stats(self):
        """
        Post processing of logs to print basic simulation statistics
        """

        df_cust, df_sys = self.return_log()

        value = np.mean(df_cust["TCWQ"]) / np.mean(df_cust["ST"])
        print('Waiting time in units of service time: {:.4f}'.format(value))
        print('')

        value = (df_cust["TSE"].iloc[-1] - np.sum(df_cust["ITS"])) / df_cust["TSE"].iloc[-1]
        print('Rho_system: system utilisation: {:.4f}'.format(value))
        value = (np.sum(df_cust["TCWQ"])/self.queue.c) / df_cust["TSE"].iloc[-1]
        print('Rho_server: server utilisation: {:.4f}'.format(value))

        value = np.sum(df_cust["ITS"]) / df_cust["TSE"].iloc[-1]
        print('P_0: probability nobody in the system: {:.4f}'.format(value))
        print('')

        value = np.mean(df_sys['c_s'])
        print('L_s: average nr of customers in the system: {}'.format(value))
        value = np.mean(df_sys['c_q'])

        print('L_q: average nr of customers in the queue: {}'.format(value))
        value = np.mean(df_cust["TCSS"])
        print('W_s: the long term average time spent in the system: {:.4f}'.format(value))
        value = np.mean(df_cust["TCWQ"])
        print('W_q: the long term average time spent in the queue: {:.4f}'.format(value))
        print('')

        value = df_cust["AT"].iloc[-1]/(len(df_cust["ST"])-1)
        print('IAT: average inter arrival time: {:.4f}'.format(value))

        value = np.sum(df_cust["ST"])/(len(df_cust["ST"]))
        print('ST: average service time: {:.4f}'.format(value))
        print('')

    def plot_system_state(self, fontsize=20):
        """
        Plot number of customers in the system and in the queue as a function of time
        """

        df_cust, df_sys = self.return_log()

        sns.set(style="white", palette="muted", color_codes=True)

        # Set up the matplotlib figure
        fig, ax = plt.subplots(figsize=(14, 7), sharex=True)
        sns.despine(left=True)

        # Plot a simple histogram with binsize determined automatically
        a = sns.lineplot(x=df_sys['t'], y=df_sys['c_s'], color="b", label="Nr. cust. in system")
        b = sns.lineplot(x=df_sys['t'], y=df_sys['c_q'], color="r", label="Nr. cust. in queue")

        # set labels
        a.set_xlabel("Time (hours)", fontsize=fontsize)
        a.set_ylabel("Nr. of customers", fontsize=fontsize)
        a.tick_params(labelsize=fontsize)

        # fig.suptitle(self.queue.kendall_notation, fontsize=fontsize)

        return fig, ax

    def plot_IAT_ST(self, fontsize=20):
        """
        Plot histograms of IAT's and ST's
        """


        df_cust, df_sys = self.return_log()

        sns.set(style="white", palette="muted", color_codes=True)

        # Set up the matplotlib figure
        fig, axes = plt.subplots(1, 2, figsize=(14, 7), sharex=True)
        sns.despine(left=True)

        # Plot a simple histogram with binsize determined automatically
        a = sns.histplot(df_cust['IAT'], kde=False, color="b", ax=axes[0])
        IAT_av = np.mean(df_cust["IAT"])
        # axes[0].plot([IAT_av, IAT_av], [0, axes[0].axis()[-1]], linewidth=3, color='k')
        axes[0].plot([IAT_av, IAT_av], [0, 1000], linewidth=3, color='k')

        b = sns.histplot(df_cust['ST'], kde=False, color="b", ax=axes[1])
        ST_av = np.mean(df_cust["ST"])
        # axes[1].plot([ST_av, ST_av], [0, axes[1].axis()[-1]], linewidth=3, color='k')
        axes[1].plot([ST_av, ST_av], [0, 1000], linewidth=3, color='k')

        # plt.setp(axes, yticks=[])
        plt.tight_layout()

        # set labels
        a.set_xlabel("Inter arrival times", fontsize=fontsize)
        a.set_ylabel("Nr. of customers", fontsize=fontsize)
        a.tick_params(labelsize=fontsize)

        b.set_xlabel("Service times", fontsize=fontsize)
        b.set_ylabel("Nr. of customers", fontsize=fontsize)
        b.tick_params(labelsize=fontsize)

        # fig.suptitle(self.queue.kendall_notation, fontsize=fontsize)

        return fig, axes




from scipy import stats
import numpy as np
import pandas as pd


class MM1:
    """
    A simple simulation method for MM1 queues
    - MM1.get_IAT_and_ST: generates lists of IAT's and ST's drawn from exponential distributions
    - MM1.calculate: values for AT, TSB, TSE, TCSS, TCWG and ITS are calculated per arrival
    - MM1.get_stats: print basic statistics based on the simulation results
    """
    def __init__(self, lam, mu, nr_arr, seed=None):
        """
        Initialization the basic time unit is hours.
        """

        # todo: consider to remove seed
        self.lam = lam  # arrivals per hour
        self.mu = mu  # departures per hour
        self.nr_arr = nr_arr  # nr of customers

        np.random.seed(seed)

    def get_IAT_and_ST(self):  # generate list of inter arrival times
        """
        Generate lists of IAT's and ST's drawn from exponential distributions.
        """

        rv_iat = stats.expon(scale=1 / self.lam)
        rv_st = stats.expon(scale=1 / self.mu)

        # generate list of inter arrival times
        IAT = rv_iat.rvs(self.nr_arr)

        # generate list of service times
        ST = rv_st.rvs(self.nr_arr)

        return IAT, ST

    def calculate(self, IAT, ST):
        """
        Values for AT, TSB, TSE, TCSS, TCWG and ITS are calculated per arrival
        """

        df_cust = pd.DataFrame()
        AT = []
        TSB = []
        TSE = []
        TCSS = []
        TCWQ = []
        ITS = []

        for i in range(len(IAT)):
            # AT
            if i == 0:
                AT.append(IAT[i])  # time starts at 0 and the first arrival arrives at 0 + IAT
            else:
                AT.append(AT[i - 1] + IAT[i])  # next arrivals start at the previous arrival AT[i-1] + IAT

            # TSB
            if i == 0:
                TSB.append(AT[i])  # first arrival the queue is empty so service begins as soon as customer arrives
            else:
                TSB.append(np.max([AT[i], TSE[i - 1]]))

            # TSE
            TSE.append(TSB[i] + ST[i])  # moment service begins plus service time

            # TCSS
            TCSS.append(TSE[i] - AT[i])  # moment of arrival until service ends

            # TCWQ
            TCWQ.append(TSB[i] - AT[i])  # moment of arrival until service begins

            # ITS
            if i == 0:
                ITS.append(IAT[i])  # the server will start idle until the first arrival
            else:
                ITS.append(np.max([AT[i] - TSE[i - 1], 0]))

        # Add lists to dataframe
        df_cust["IAT"] = IAT
        df_cust["ST"] = ST
        df_cust["AT"] = AT
        df_cust["TSB"] = TSB
        df_cust["TSE"] = TSE
        df_cust["TCSS"] = TCSS
        df_cust["TCWQ"] = TCWQ
        df_cust["ITS"] = ITS

        return df_cust

    def get_stats(self, df_cust):
        """
        Generate lists of IAT's and ST's drawn from exponential distributions.
        """

        value = np.mean(df_cust["TCWQ"]) / np.mean(df_cust["ST"])
        print('Waiting time over service time: {:.4f}'.format(value))
        print('')

        value = (df_cust["TSE"].iloc[-1] - np.sum(df_cust["ITS"])) / df_cust["TSE"].iloc[-1]
        print('Rho: system utilisation: {:.4f}'.format(value))
        print('')

        value = np.sum(df_cust["ITS"]) / df_cust["TSE"].iloc[-1]
        print('P_0: probability nobody in the system: {:.4f}'.format(value))

        value = np.mean(df_cust["TCSS"])
        print('W_s: the long term average time spent in the system: {:.4f}'.format(value))

        value = np.mean(df_cust["TCWQ"])
        print('W_q: the long term average time spent in the queue: {:.4f}'.format(value))

        value = df_cust["AT"].iloc[-1]/(len(df_cust["ST"])-1)
        print('IAT: average inter arrival time: {:.4f}'.format(value))

        value = np.sum(df_cust["ST"])/(len(df_cust["ST"]))
        print('ST: average service time: {:.4f}'.format(value))
        print('')



"""Top-level package for queueing."""

__author__ = "Mark van Koningsveld and Joris den Uijl"
__email__ = "m.vankoningsveld@tudelft.nl"
__version__ = "v0.5.1"

from .arrival_process import ArrivalProcess
from .customer import Customer
from .mm1 import MM1
from .mt_engine import worker, Task
from .queue import Queue
from .service_process import ServiceProcess
from .simulation import Simulation





class Customer:
    """
    Customer class for use in the OpenQTSim package
    """

    def __init__(self, Env, Sim):
        """
        Initialization
        """

        # self.arrival = arrival
        self.Env = Env
        self.Sim = Sim

        Sim.customer_nr += 1
        self.customer_nr = Sim.customer_nr

    def move(self, IAT, AT):
        """"
        Method to move Customer through the system
        """
        # request access to server
        self.Sim.c_s += 1
        self.Sim.c_q += 1
        t_req = AT
        c_s_req = self.Sim.c_s
        c_q_req = self.Sim.c_q

        server = yield self.Env.servers.get()

        self.Sim.c_q -= 1
        t_yield = self.Env.now - self.Env.epoch
        c_s_yield = self.Sim.c_s
        c_q_yield = self.Sim.c_q

        if t_req != t_yield:  # only log when someone is actually waiting to be served
            self.Sim.log_system_state(t_req, c_s_req, c_q_req)
            self.Sim.log_system_state(t_yield, c_s_yield, c_q_yield)
        else:  # otherwise log the yield moment
            self.Sim.log_system_state(t_yield, c_s_yield, c_q_yield)

        # register if the server was idle
        ITS = self.Env.now - self.Env.server_info[server.id]['last_active']

        # register which server was used
        s_id = server.id

        # determine TSB
        TSB = self.Env.now - self.Env.epoch

        # get ST
        ST = self.Sim.queue.S.get_ST(server, customer_nr=self.customer_nr)

        # move time ST forward
        yield self.Env.timeout(ST)

        # determine TSE
        TSE = self.Env.now - self.Env.epoch

        self.Sim.c_s -= 1
        if self.Sim.c_q == 0:
            self.Sim.log_system_state(TSE, self.Sim.c_s, self.Sim.c_q)
        # Todo: when a customer leaves the system while somebody is still in the queue, you get a double logging
        #  (check how this works for more than 1 server)

        # update server_info when server was last active
        self.Env.server_info.update({server.id: {'last_active': self.Env.now}})

        # release server when done
        yield self.Env.servers.put(server)

        # add customer info to log
        self.Sim.log_customer_state(self.customer_nr, IAT, AT, ST, TSB, TSE, ITS, s_id)





class ServiceProcess:
    """
    Server process class for use in the OpenQTSim package
    """

    def __init__(self, symbol='M', srv_rate=9):
        """
        symbol: symbol of the process (M, E_k, etc.)
        srv_rate: services per hour
        """

        self.symbol = symbol
        self.srv_rate = srv_rate

    def get_ST(self, server, customer_nr=[]):
        """
        Return the inter arrival time based on the inter arrival time distribution or deterministic list
        """

        if self.symbol == "M" or self.symbol[0] == "E":
            return server.service_distribution.rvs()

        elif self.symbol == "D":
            return server.service_distribution.loc[customer_nr, ['ST']].item()



import openqtsim
import numpy as np
from collections import namedtuple

Task = namedtuple('Task', 'A, S, c, nr_arr, lam, mu')


def worker(task:Task):
    # calculate the appropriate service rate per server
    srv_rate = task.mu/task.c

    # create Arrival and Service processes and specify the number of servers
    A = openqtsim.ArrivalProcess(task.A, arr_rate=task.lam)
    S = openqtsim.ServiceProcess(task.S, srv_rate=srv_rate)
    c = task.c

    # create the queue object
    q = openqtsim.Queue(A, S, c)

    # use the queue object to create a simulation object and run simulation with the specified number of arrivals
    sim = openqtsim.Simulation(q)
    sim.run(task.nr_arr)

    # retrieve the logs (df1: customer log, df2: system log)
    df1, df2 = sim.return_log()

    # use the customer log to determine the average waiting time as a factor of service time 
    factor = np.mean(df1["TCWQ"]) / np.mean(df1["ST"])

    return factor



"""
Setup file for hydraulic_infrastructure_realisation.
Use setup.cfg to configure your project.

This file was generated with PyScaffold 3.1.
PyScaffold helps you to put up the scaffold of your new Python project.
Learn more under: https://pyscaffold.org/
"""
import sys

from pkg_resources import require, VersionConflict
from setuptools import setup, find_packages

try:
    require("setuptools>=38.3")
except VersionConflict:
    print("Error: version of setuptools is too old (<38.3)!")
    sys.exit(1)

requires = [
    "matplotlib",
    "pandas",
    "numpy",
    "seaborn",
    "simpy",
    "scipy",
    "sphinx_rtd_theme",
]

setup_requirements = [
    "pytest-runner",
]

tests_require = [
    "pytest",
    "pytest-cov",
]

with open("README.md", "r") as des:
    long_description = des.read()

setup(
    author="Mark van Koningsveld and Joris den Uijl",
    author_email="m.vankoningsveld@tudelft.nl",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="OpenQTSim facilitates discrete event simulation of queues with a Kendall notation.",
    install_requires=requires,
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="Queueing Theory",
    name="openqtsim",
    packages=find_packages(include=["openqtsim"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=tests_require,
    url="https://github.com/TUDelft-CITG/OpenQTSim",
    version="v0.5.1",
    zip_safe=False,
)





class ArrivalProcess:
    """
    Arrival process class for use in the OpenQTSim package
    """

    def __init__(self, symbol='M', arr_rate=8):
        """
        symbol: symbol of the process (M, E_k, etc.)
        arr_rate: arrivals per hour
        """

        self.symbol = symbol
        self.arr_rate = arr_rate

    def get_IAT(self, customer_nr=[]):
        """
        Return the inter arrival time based on the inter arrival time distribution or deterministic list
        """

        if self.symbol == "M" or self.symbol[0] == "E":
            return self.arrival_distribution.rvs()

        elif self.symbol == "D":
            return self.arrival_distribution.loc[customer_nr, ['IAT']].item()



import numpy as np
import openqtsim

"""
"""


def test_lookup_table_1():
    queue = openqtsim.Queue()

    utilisation = 0.9
    nr_of_servers = 1

    factor = queue.occupancy_to_waitingfactor(utilisation, nr_of_servers)
    answer = 8.998295524474928

    np.testing.assert_almost_equal(factor, answer)

def test_lookup_table_2():
    queue = openqtsim.Queue()

    factor = 0.0870
    nr_of_servers = 4

    factor = queue.waitingfactor_to_occupancy(factor, nr_of_servers)
    answer = 0.493062176851717

    np.testing.assert_almost_equal(factor, answer)
